@@ -1,6 +1,7 @@
 import os
 from pathlib import Path
 import numpy as np
+import pandas as pd
 from keras.preprocessing.image import load_img, save_img, img_to_array
 from keras.applications.imagenet_utils import preprocess_input
 from keras.preprocessing import image
@@ -8,6 +9,14 @@ import cv2
 from pathlib import Path
 import gdown
 import hashlib
+import math
+from PIL import Image
+
+def distance(a, b):
+	x1 = a[0]; y1 = a[1]
+	x2 = b[0]; y2 = b[1]
+	
+	return math.sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1)))
 
 def findFileHash(file):
 	BLOCK_SIZE = 65536 # The size of each read from the file
@@ -95,7 +104,7 @@ def findThreshold(model_name, distance_metric):
 		elif distance_metric == 'euclidean':
 			threshold = 64
 		elif distance_metric == 'euclidean_l2':
-			threshold = 0.69
+			threshold = 0.64
 	
 	return threshold
 
@@ -108,41 +117,128 @@ def detectFace(image_path, target_size=(224, 224), grayscale = False):
 	for folder in folders[1:]:
 		path = path + "/" + folder
 
-	detector_path = path+"/data/haarcascade_frontalface_default.xml"
+	face_detector_path = path+"/data/haarcascade_frontalface_default.xml"
+	eye_detector_path = path+"/data/haarcascade_eye.xml"
 	
-	if os.path.isfile(detector_path) != True:
-		raise ValueError("Confirm that opencv is installed on your environment! Expected path ",detector_path," violated.")
+	if os.path.isfile(face_detector_path) != True:
+		raise ValueError("Confirm that opencv is installed on your environment! Expected path ",face_detector_path," violated.")
 	
 	#--------------------------------
 	
-	detector = cv2.CascadeClassifier(detector_path)
+	face_detector = cv2.CascadeClassifier(face_detector_path)
+	eye_detector = cv2.CascadeClassifier(eye_detector_path)
 	
-	if grayscale != True:
-		img = cv2.imread(image_path)
-	else: #gray scale
-		img = cv2.imread(image_path, 0)
+	img = cv2.imread(image_path)
 	
-	faces = detector.detectMultiScale(img, 1.3, 5)
+	img_raw = img.copy()
+	
+	#--------------------------------
+	
+	faces = face_detector.detectMultiScale(img, 1.3, 5)
 	
 	#print("found faces in ",image_path," is ",len(faces))
 	
 	if len(faces) > 0:
 		x,y,w,h = faces[0]
 		detected_face = img[int(y):int(y+h), int(x):int(x+w)]
+		detected_face_gray = cv2.cvtColor(detected_face, cv2.COLOR_BGR2GRAY)
+		
+		#---------------------------
+		#face alignment
+		
+		eyes = eye_detector.detectMultiScale(detected_face_gray)
+		
+		if len(eyes) >= 2:
+			#find the largest 2 eye
+			base_eyes = eyes[:, 2]
+			
+			items = []
+			for i in range(0, len(base_eyes)):
+				item = (base_eyes[i], i)
+				items.append(item)
+			
+			df = pd.DataFrame(items, columns = ["length", "idx"]).sort_values(by=['length'], ascending=False)
+			
+			eyes = eyes[df.idx.values[0:2]]
+			
+			#-----------------------
+			#decide left and right eye
+			
+			eye_1 = eyes[0]; eye_2 = eyes[1]
+			
+			if eye_1[0] < eye_2[0]:
+				left_eye = eye_1
+				right_eye = eye_2
+			else:
+				left_eye = eye_2
+				right_eye = eye_1
+			
+			#-----------------------
+			#find center of eyes
+			
+			left_eye_center = (int(left_eye[0] + (left_eye[2] / 2)), int(left_eye[1] + (left_eye[3] / 2)))
+			left_eye_x = left_eye_center[0]; left_eye_y = left_eye_center[1]
+			
+			right_eye_center = (int(right_eye[0] + (right_eye[2]/2)), int(right_eye[1] + (right_eye[3]/2)))
+			right_eye_x = right_eye_center[0]; right_eye_y = right_eye_center[1]
+			
+			#-----------------------
+			#find rotation direction
+				
+			if left_eye_y > right_eye_y:
+				point_3rd = (right_eye_x, left_eye_y)
+				direction = -1 #rotate same direction to clock
+			else:
+				point_3rd = (left_eye_x, right_eye_y)
+				direction = 1 #rotate inverse direction of clock
+			
+			#-----------------------
+			#find length of triangle edges
+			
+			a = distance(left_eye_center, point_3rd)
+			b = distance(right_eye_center, point_3rd)
+			c = distance(right_eye_center, left_eye_center)
+			
+			#-----------------------
+			#apply cosine rule
+			
+			cos_a = (b*b + c*c - a*a)/(2*b*c)
+			angle = np.arccos(cos_a) #angle in radian
+			angle = (angle * 180) / math.pi #radian to degree
+			
+			#-----------------------
+			#rotate base image
+			
+			if direction == -1:
+				angle = 90 - angle
+			
+			img = Image.fromarray(img_raw)
+			img = np.array(img.rotate(direction * angle))
+			
+			#you recover the base image and face detection disappeared. apply again.
+			faces = face_detector.detectMultiScale(img, 1.3, 5)
+			if len(faces) > 0:
+				x,y,w,h = faces[0]
+				detected_face = img[int(y):int(y+h), int(x):int(x+w)]
+			
+			#-----------------------
+		
+		#face alignment block end
+		#---------------------------
+		
+		#face alignment block needs colorful images. that's why, converting to gray scale logic moved to here.
+		if grayscale == True:
+			detected_face = cv2.cvtColor(detected_face, cv2.COLOR_BGR2GRAY)
+		
 		detected_face = cv2.resize(detected_face, target_size)
 		
 		img_pixels = image.img_to_array(detected_face)
 		img_pixels = np.expand_dims(img_pixels, axis = 0)
 		
-		if True:
-			#normalize input in [0, 1]
-			img_pixels /= 255 
-		else:
-			#normalize input in [-1, +1]
-			img_pixels /= 127.5
-			img_pixels -= 1
+		#normalize input in [0, 1]
+		img_pixels /= 255
 		
 		return img_pixels
 		
 	else:
-		raise ValueError("Face could not be detected in ", image_path,". Please confirm that the picture is a face photo.")
+		raise ValueError("Face could not be detected in ", image_path,". Please confirm that the picture is a face photo.")
\ No newline at end of file
